import numpy as np
from typing import Optional
import h5py
from os.path import basename

from emdfile.tqdmnd import tqdmnd
from emdfile.classes.tree import Node
from emdfile.classes.pointlist import PointList

class PointListArray(Node):
    """
    An 2D array of PointLists which share common coordinates.
    """

    _emd_group_type = "pointlistarray"

    def __init__(
        self,
        dtype,
        shape,
        name: Optional[str] = 'pointlistarray',
        ):
        """
		Creates an empty PointListArray.

        Args:
            dtype: the dtype of the numpy structured arrays which will comprise
                the data of each PointList
            shape (2-tuple of ints): the shape of the array of PointLists
            name (str): a name for the PointListArray

        Returns:
            a PointListArray instance
        """
        assert len(shape) == 2, "Shape must have length 2."

        super().__init__()

        self.name = name
        self.shape = shape

        self.dtype = np.dtype(dtype)
        self.fields = self.dtype.names
        if self.fields is not None:
            self.types = tuple([self.dtype.fields[f][0] for f in self.fields])
        else:
            self.fields = ('',)
            self.types = (dtype,)


        # Populate with empty PointLists
        self._pointlists = [[PointList(data=np.zeros(0,dtype=self.dtype), name=f"{i},{j}")
                             for j in range(self.shape[1])] for i in range(self.shape[0])]


    ## get/set pointlists

    def __getitem__(self, tup):
        l = len(tup) if isinstance(tup,tuple) else 1
        assert(l==2), f"Expected 2 slice values, recieved {l}"
        return self.get_pointlist(tup[0],tup[1])

    def __setitem__(self, tup, pointlist):
        l = len(tup) if isinstance(tup,tuple) else 1
        assert(l==2), f"Expected 2 slice values, recieved {l}"
        assert(pointlist.fields == self.fields), "fields must match"
        self._pointlists[tup[0]][tup[1]] = pointlist

    def get_pointlist(self, i, j, name=None):
        """
        Returns the pointlist at i,j
        """
        pl = self._pointlists[i][j]
        if name is not None:
            pl = pl.copy(name=name)
        return pl



    ## Make copies

    def copy(self, name=''):
        """
        Returns a copy of itself.
        """
        new_pla = PointListArray(
            dtype=self.dtype,
            shape=self.shape,
            name=name)

        for i in range(new_pla.shape[0]):
            for j in range(new_pla.shape[1]):
                pl = new_pla.get_pointlist(i,j)
                pl.add(np.copy(self.get_pointlist(i,j).data))

        for k,v in self.metadata.items():
            new_pla.metadata = v.copy(name=k)

        return new_pla

    def add_fields(self, new_fields, name=''):
        """
        Creates a copy of the PointListArray, but with additional fields given
        by new_fields.

        Args:
            new_fields: a list of 2-tuples, ('name', dtype)
            name: a name for the new pointlist
        """
        dtype = []
        for f,t in zip(self.fields,self.types):
            dtype.append((f,t))
        for f,t in new_fields:
            dtype.append((f,t))

        new_pla = PointListArray(
            dtype=dtype,
            shape=self.shape,
            name=name)

        for i in range(new_pla.shape[0]):
            for j in range(new_pla.shape[1]):
                # Copy old data into a new structured array
                pl_old = self.get_pointlist(i,j)
                data = np.zeros(pl_old.length, np.dtype(dtype))
                for f in self.fields:
                    data[f] = np.copy(pl_old.data[f])

                # Write into new pointlist
                pl_new = new_pla.get_pointlist(i,j)
                pl_new.add(data)

        return new_pla




    ## Representation to standard output
    def __repr__(self):

        space = ' '*len(self.__class__.__name__)+'  '
        string = f"{self.__class__.__name__}( A shape {self.shape} PointListArray called '{self.name}',"
        string += "\n"+space+f"with {len(self.fields)} fields:"
        string += "\n"
        space2 = max([len(field) for field in self.fields])+3
        for f,t in zip(self.fields,self.types):
            string += "\n"+space+f"{f}{(space2-len(f))*' '}({str(t)})"
        string += "\n)"

        return string



    # HDF5 i/o

    # write
    def to_h5(self,group):
        """
        Takes an h5py Group instance and creates a subgroup containing
        this PointListArray, tags indicating its EMD type and Python class,
        and the pointlistarray's data and metadata.

        Accepts:
            group (h5py Group)

        Returns:
            (h5py Group) the new pointlistarray's group
        """
        # Construct group and add metadata
        grp = Node.to_h5(self,group)

        # Add metadata
        dtype = h5py.special_dtype(vlen=self.dtype)
        dset = grp.create_dataset(
            "data",
            self.shape,
            dtype
        )

        # Add data
        for (i,j) in tqdmnd(dset.shape[0],dset.shape[1]):
            dset[i,j] = self[i,j].data

        # Return
        return grp


    # read
    @classmethod
    def _get_constructor_args(cls,group):
        """
        Returns a dictionary of args/values to pass to the class constructor
        """
        # Get the DataSet
        dset = group['data']
        dtype = h5py.check_vlen_dtype( dset.dtype )
        shape = dset.shape

        # make args dictionary and return
        return {
            'dtype' : dtype,
            'shape' : shape,
            'name' : basename(group.name)
        }



        # Add metadata
        _read_metadata(pla, group)

        return pla

    def _populate_instance(self,group):
        """
        Accepts an already extant class self, and populates it with the data from h5py Group `group`
        """
        # Find the data and shape
        dset = group['data']
        shape = self.shape
        # Add data
        for (i,j) in tqdmnd(shape[0],shape[1],desc="Reading PointListArray",unit="PointList"):
            try:
                self[i,j].add(dset[i,j])
            except ValueError:
                pass
        return self






import numpy as np
from emdfile import (
    Node,
    Metadata,
    Array,
    PointList,
    PointListArray
)

from os.path import join,exists
from os import remove
from numpy import array_equal

from emdfile import _TESTPATH
from emdfile import save, read

# Set paths
dirpath = _TESTPATH
testpath = join(dirpath,"test_base_classes.h5")





class TestArray():


    def test_instantiation(self):

        shape = (3,4,5)
        d = np.arange(np.prod(shape)).reshape(shape)

        ar = Array(
            data = d
        )
        assert(isinstance(ar, Array))


    def test_Array_complex_readwrite(self):

        # make array
        d = np.exp(1j*np.linspace(0,2*np.pi)).astype(np.complex64)
        ar = Array(
            data = d
        )

        # save, read, check
        save(testpath,ar,mode='o')
        ar2 = read(testpath)
        assert(array_equal(ar.data,ar2.data))


    def test_Array_stackarray(self):

        # make array
        x = Array(np.zeros((3,5,6)),slicelabels=['a','b','c'])

        # save
        save(testpath,x,mode='o')

        # read
        y = read(testpath)

        # compare
        assert(x.depth == y.depth)
        assert(x.rank == y.rank)
        assert(x.shape == y.shape)
        assert(x.slicelabels == y.slicelabels)




    def test_Array_dimvects(self):

        dims = (
            [0,0.5],
            None,
            [0,3,4,5,9,100]
        )

        # make array
        shape = (4,5,6,8)
        data_shape = tuple([3]+list(shape))
        d = np.arange(np.prod(data_shape)).reshape(data_shape)
        ar = Array(
            data = d,
            dims = dims,
            slicelabels = ['a','b','c']
        )


        # check that...

        # ...the last (3'rd) dim instantiated in integer pixels
        assert(np.array_equal(
            ar.dim(3),
            np.arange(shape[3])
        ))
        # ...the 2nd dim instantiated as written
        assert(np.array_equal(
            ar.dim(2),
            np.array(dims[2])
        ))
        # ...the 1st dim expanded correctly to pixels
        assert(np.array_equal(
            ar.dim(1),
            np.arange(shape[1])
        ))
        # ...the 0th dim expanded correctly
        assert(np.array_equal(
            ar.dim(0),
            np.arange(shape[0])*0.5
        ))

        # ...dim names are correct
        assert(ar.dim_names[0] == 'dim0')
        assert(ar.dim_names[1] == 'dim1')
        assert(ar.dim_names[2] == 'dim2')
        assert(ar.dim_names[3] == 'dim3')

        # ... the dim units are correct
        assert(ar.dim_units[0] == 'unknown')
        assert(ar.dim_units[1] == 'pixels')
        assert(ar.dim_units[2] == 'unknown')
        assert(ar.dim_units[3] == 'pixels')

        # modify a dim and then check it
        new_dim = np.array([-3,-1.1,0.8,1.12,2.5])
        ar.set_dim(
            1,
            new_dim,
            units = 'cows',
            name = 'fields'
        )
        assert(np.array_equal(
            ar.dim(1),
            new_dim
        ))
        assert(ar.dim_units[1] == 'cows')
        assert(ar.dim_names[1] == 'fields')



        # save, read
        save(testpath,ar,mode='o')
        ar2 = read(testpath)

        # check...

        # data
        assert(array_equal(ar.data,ar2.data))

        # dims
        assert(array_equal(ar.dim(0), ar2.dim(0)))
        assert(array_equal(ar.dim(1), ar2.dim(1)))
        assert(array_equal(ar.dim(2), ar2.dim(2)))
        assert(array_equal(ar.dim(3), ar2.dim(3)))

        # dim names
        assert(ar.dim_names[0] == ar2.dim_names[0])
        assert(ar.dim_names[1] == ar2.dim_names[1])
        assert(ar.dim_names[2] == ar2.dim_names[2])
        assert(ar.dim_names[3] == ar2.dim_names[3])

        # dim units
        assert(ar.dim_units[0] == ar2.dim_units[0])
        assert(ar.dim_units[1] == ar2.dim_units[1])
        assert(ar.dim_units[2] == ar2.dim_units[2])
        assert(ar.dim_units[3] == ar2.dim_units[3])

        # slices
        assert(array_equal( ar['a'].data, ar2['a'].data ))
        assert(array_equal( ar['b'].data, ar2['b'].data ))
        assert(array_equal( ar['c'].data, ar2['c'].data ))






class TestPointList():


    def setup_method(self):
        """ instantiate a PointList
        """
        dtype = [('x',np.int32),('y',np.float32)]

        pl = PointList(
            data = np.zeros(5,dtype=dtype)
        )
        pl['x'][:] = np.arange(5)
        self.pl = pl
        pass


    def test_instantiation(self):
        assert isinstance(self.pl, PointList)
        pass


    def test_PointList_readwrite(self):

        # save, read, check
        save(testpath,self.pl,mode='o')
        pl2 = read(testpath)
        assert(array_equal(pl2.data,self.pl.data))


    def test_PointList_addition(self):

        # make a structured array of the same dtype
        dtype = self.pl.dtype
        new_data = np.ones(3,dtype=dtype)

        # add, resulting in a new pointlist
        # length should be len(old)+len(new)
        new_pointlist = self.pl + new_data
        assert(len(new_pointlist) == 8)

        # the old pointlist should be unmodified
        assert(len(self.pl) == 5)

        # alternatively if we call the .add method,
        # the old pointlist is modified
        self.pl.add(new_data)
        assert(len(self.pl) == 8)

        # or if we perform a new assignment, the old
        # pointlist is modified
        self.pl += new_pointlist
        assert(len(self.pl) == 16)


    def test_simple_vector_PointList(self):

        # make a PointList with a *non*-structured array
        pl = PointList(
            name = 'test_simple_pointlist',
            data = np.arange(5).astype(np.float32)
        )

        # type check
        assert(isinstance(pl,PointList))
        assert(pl.fields == ('',))
        assert(pl._types == (np.float32,))

        # make data to add
        new_data = np.arange(5,8).astype(np.float32)

        # add, resulting in a new pointlist
        # length should be len(old)+len(new)
        new_pointlist = pl + new_data
        assert(len(new_pointlist) == 8)

        # the old pointlist should be unmodified
        assert(len(pl) == 5)

        # alternatively if we call the .add method,
        # the old pointlist is modified
        pl.add(new_data)
        assert(len(pl) == 8)

        # or if we perform a new assignment, the old
        # pointlist is modified
        pl += new_pointlist
        assert(len(pl) == 16)



def test_Node():

    # Root class instances should:
    # - have a name
    # - have a Tree
    # - know how to read/write to/from h5

    root = Node()
    assert(isinstance(root,Node))
    ##;passert(root.name == 'root')
    ##;passert(isinstance(root.tree, Tree))

    # h5io


def test_Metadata():

    # Metadata class instances should:
    # - TODO

    metadata = Metadata()
    assert(isinstance(metadata,Metadata))







def test_PointList():

    # PointList class instances should:
    # - TODO

    dtype = [
        ('x',int),
        ('y',float)
    ]
    data = np.zeros(10,dtype=dtype)
    pointlist = PointList(
        data=data
    )
    assert(isinstance(pointlist,PointList))

def test_PointListArray():

    # PointListArray class instance should:
    # - TODO

    dtype = [
        ('x',int),
        ('y',float)
    ]
    shape = (5,5)
    pointlistarray = PointListArray(
        dtype = dtype,
        shape = shape
    )
    assert(isinstance(pointlistarray,PointListArray))











"""
A wrapper for tqdm (https://github.com/tqdm/tqdm) that supports
multidimensional iterators. Iterates over the Cartesian product
of the input iterators. Integers in the input are automagically
conveted to range(i) for convenience.
Examples:
for x,y in tqdmnd(range(20),range(10)):
    sleep(0.1)

for x,y in tqdmnd(20,10):
    sleep(0.1)
"""

from tqdm import tqdm

from itertools import product
from functools import reduce
from operator import mul
from numpy import integer

from collections.abc import Iterator


class nditer(Iterator):
    def __init__(self, *args):
        if len(args) > 1:
            self._it = product(*args)
        else:
            self._it = args[0]
        self._l = reduce(mul, [a.__len__() for a in args])

    def __iter__(self):
        return self._it.__iter__()

    def __next__(self):
        return self._it.__next__()

    def __len__(self):
        return self._l


def tqdmnd(*args, **kwargs):
    """
    An N-dimensional extension of tqdm providing an iterator and
    progress bar over the product of multiple iterators.

    Example Usage:

        >>> for x,y in tqdmnd(5,6):
        >>>     <expression>

    is equivalent to

        >>> for x in range(5):
        >>>     for y in range(6):
        >>>         <expression>

    with a tqdmnd-style progress bar printed to standard output.

    Accepts:
        *args: Any number of integers or iterators. Each integer N
            is converted to a `range(N)` iterator. Then a loop is
            constructed from the Cartesian product of all iterables.
        **kwargs: keyword arguments passed through directly to tqdm.
            Full details are available at https://tqdm.github.io
            A few useful ones:
                disable (bool): if True, hide the progress bar
                keep (bool): if True, delete the progress bar after completion
                unit (str): unit name for the display of iteration speed
                unit_scale (bool): whether to scale the displayed units and add
                    SI prefixes
                desc (str): message displayed in front of the progress bar

    Returns:
        At each iteration, a tuple of indices is returned, corresponding to the
        values of each input iterator (in the same order as the inputs).
    """
    r = [range(i) if isinstance(i, (int, integer)) else i for i in args]
    return tqdm(nditer(*r), **kwargs)



# Reader functions
# for native and non-native file types

import h5py
import pathlib
from os.path import exists, splitext, basename, dirname, join
from typing import Union, Optional
import warnings


# Classes
from emdfile.classes import (
    Root,
    Node,
    Metadata
)
from emdfile.classes.utils import (
    _get_class,
    EMD_data_group_types
)





# read utilities

def _get_EMD_rootgroups(filepath):
    """
    Returns a list of root groups in an EMD 1.0 file.
    """
    rootgroups = []
    with h5py.File(filepath,'r') as f:
        for key in f.keys():
            if 'emd_group_type' in f[key].attrs:
                if f[key].attrs['emd_group_type'] == 'root':
                    rootgroups.append(key)
    return rootgroups

def _is_EMD_file(filepath):
    """
    Returns True iff filepath points to a valid EMD 1.0 file.
    """
    # check for the 'emd_group_type'='file' attribute
    with h5py.File(filepath,'r') as f:
        try:
            assert('emd_group_type' in f.attrs.keys())
            assert('version_major' in f.attrs.keys())
            assert('version_minor' in f.attrs.keys())
            assert(f.attrs['emd_group_type'] == 'file')
            assert(f.attrs['version_major'] == 1)
            assert(f.attrs['version_minor'] == 0)
        except AssertionError:
            return False
    rootgroups = _get_EMD_rootgroups(filepath)
    if len(rootgroups)>0:
        return True
    else:
        return False

def _get_EMD_version(filepath, rootgroup=None):
    """
    Returns the version (major,minor,release) of an EMD file.
    """
    assert(_is_EMD_file(filepath)), "Error: not recognized as an EMD file"
    with h5py.File(filepath,'r') as f:
        v_major = int(f.attrs['version_major'])
        v_minor = int(f.attrs['version_minor'])
        if 'version_release' in f.attrs.keys():
            v_release = int(f.attrs['version_release'])
        else:
            v_release = 0
        return v_major, v_minor, v_release

def _get_UUID(filepath):
    """
    Returns the UUID of an EMD file, or if unavailable returns -1.
    """
    assert(_is_EMD_file(filepath)), "Error: not recognized as an EMD file"
    with h5py.File(filepath,'r') as f:
        if 'UUID' in f.attrs:
            return f.attrs['UUID']
    return -1

def _version_is_geq(current,minimum):
    """
    Returns True iff current version (major,minor,release) is greater than or equal to minimum."
    """
    if current[0]>minimum[0]:
        return True
    elif current[0]==minimum[0]:
        if current[1]>minimum[1]:
            return True
        elif current[1]==minimum[1]:
            if current[2]>=minimum[2]:
                return True
        else:
            return False
    else:
        return False






########## EMD 1.0+ reader ##########

def read(
    filepath,
    emdpath: Optional[str] = None,
    tree: Optional[Union[bool,str]] = True,
    **legacy_options,
    ):
    """
    File reader for EMD 1.0+ files.

    Args:
        filepath (str or Path): the file path
        emdpath (str): path to the node in an EMD object tree to read
            from. May be a root node or some downstream node. Use '/'
            delimiters between node names. If emdpath is None, checks to
            see how many root nodes are present. If there is one, loads
            this tree. If there are several, returns a list of the root names.
        tree (True or False or 'branch'): indicates what data should be loaded,
            relative to the node specified by `emdpath`. If set to `False`,
            only data/metadata in the specified node is loaded, plus any
            root metadata. If set to `True`, loads that node plus the
            subtree of data objects it contains (and their metadata, and
            the root metadata). If set to `'branch'`, loads the branch
            under this node as above, but does not load the node itself.
            If `emdpath` points to a root node, setting `tree` to `'branch`'
            or `True` are equivalent - both return the whole data tree.

    Returns:
        (Root) returns a Root instance containing (1) any root metadata from
            the EMD tree loaded from, and (2) a tree of one or more pieces
            of data/metadata
    """
    # parse filepath, verify EMD file
    er1 = f"filepath must be a string or Path, not {type(filepath)}"
    er2 = f"specified filepath '{filepath}' was not found on the filesystem"
    er3 = f"{filepath} isn't a valid EMD 1.0 file."
    assert(isinstance(filepath, (str,pathlib.Path) )), er1
    assert(exists(filepath)), er2
    assert(_is_EMD_file(filepath)), er3


    # get/check version
    v = _get_EMD_version(filepath)


    # determine `emdpath` if it was left as None
    if emdpath is None:
        rootgroups = _get_EMD_rootgroups(filepath)
        if len(rootgroups) == 0:
            raise Exception("No root groups found! This error should never occur! (You're amazing! You've broken the basic laws of logic, reason, and thermodynamics itself!)")
        elif len(rootgroups) == 1:
            emdpath = rootgroups[0]
        else:
            print("Multiple root groups detected - returning root names. Please specify the `emdpath` argument. Returning the list of rootgroups.")
            return rootgroups



    # parse the root and tree paths
    p = emdpath.split('/')
    if '' in p:
        p.remove('')
    rootpath = p[0]
    treepath = '/'.join(p[1:])



    # Open the h5 file
    with h5py.File(filepath,'r') as f:

        # Find the root group
        assert(rootpath in f.keys()), f"Error: root group {rootpath} not found"
        rootgroup = f[rootpath]

        # Find the node of interest
        group_names = treepath.split('/')
        nodegroup = rootgroup
        if len(group_names)==1 and group_names[0]=='':
            pass
        else:
            for name in group_names:
                assert(name in nodegroup.keys()), f"Error: group {name} not found in group {nodegroup.name}"
                nodegroup = nodegroup[name]

        # Read the root
        root = Root.from_h5(rootgroup)

        # if this is all that was requested, return
        if nodegroup is rootgroup and tree is False:
                return root

        # Read...

        # ...if the whole tree was requested
        if nodegroup is rootgroup and tree in (True,'branch'):
            # build the tree
            n = _populate_tree(root,rootgroup)
            # return...
            if n == 1:
                # ...if there's one node, return it
                key = list(root._branch.keys())[0]
                node = root.tree(key)
            elif n == 0 and len(root.metadata) == 1:
                # ...if there's no nodes and one dictionary,
                # return it
                key = list(root.metadata.keys())[0]
                node = root.metadata[key]
            else:
                # ...otherwise, return the root
                node = root

        # ...if a single node was requested
        elif tree is False:
            # read the node
            node = _read_single_node(nodegroup)
            # build the tree and return
            root.force_add_to_tree(node)

        # ...if a branch was requested
        elif tree is True:
            # read source node and add to tree
            node = _read_single_node(nodegroup)
            root.force_add_to_tree(node)
            # build the tree
            _populate_tree(node,nodegroup)

        # ...if `tree == 'branch'`
        else:
            # build the tree
            _populate_tree(root,nodegroup)
            node = root

    # Return
    return node





# group / tree reading utilities

def _read_single_node(grp):
    """
    Determines the class type of the h5py Group `grp`, then
    instantiates and returns an instance of the class with
    this group's data and metadata
    """
    __class__ = _get_class(grp)
    data = __class__.from_h5(grp)
    return data

def _populate_tree(node,group,count=0):
    """
    `node` is a Node and `group` is its parallel h5py Group.
    Reads the tree underneath this nodegroup in the h5 file and adds it
    to the runtime tree underneath this node. Does *not* read `group`
    itself - this function grafts everything underneath `group` onto node

    Returns the number of new nodes added to the tree
    """
    keys = [k for k in group.keys() if isinstance(group[k],h5py.Group)]
    keys = [k for k in keys if 'emd_group_type' in group[k].attrs.keys()]
    keys = [k for k in keys if group[k].attrs['emd_group_type'] in \
        EMD_data_group_types]

    for key in keys:
        new_node = _read_single_node(group[key])
        count += 1
        node.force_add_to_tree(new_node)
        _populate_tree(
            new_node,
            group[key],
            count = count
        )
    return count







# Print the HDF5 filetree to screen

def print_h5_tree(filepath, show_metadata=False):
    """
    Prints the contents of an h5 file from a filepath.
    """
    with h5py.File(filepath,'r') as f:
        print('/')
        _print_h5pyFile_tree(f, show_metadata=show_metadata)
        print('\n')

def _print_h5pyFile_tree(f, tablevel=0, linelevels=[], show_metadata=False):
    """
    Prints the contents of an h5 file from an open h5py File instance.
    """
    if tablevel not in linelevels:
        linelevels.append(tablevel)
    keys = [k for k in f.keys() if isinstance(f[k],h5py.Group)]
    if not show_metadata:
        keys = [k for k in keys if k != 'metadatabundle']
    N = len(keys)
    for i,k in enumerate(keys):
        string = ''
        string += '|' if 0 in linelevels else ' '
        for idx in range(tablevel):
            l = '|' if idx+1 in linelevels else ' '
            string += '   '+l
        print(string+'---'+k)
        if i == N-1:
            linelevels.remove(tablevel)
        _print_h5pyFile_tree(
            f[k],
            tablevel=tablevel+1,
            linelevels=linelevels,
            show_metadata=show_metadata)

    pass



def _read_metadata(
    group,
    name
    ):
    """
    Returns a Metadata instance called name stored in the EMD node at group.
    Returns False otherwise.
    """
    try:
        grp_metadata = group['metadatabundle']
    except KeyError:
        return False
    try:
        metadata = Metadata.from_h5(grp_metadata[name])
        return metadata
    except KeyError:
        return False








from emdfile import Custom, Metadata, _read_metadata
import numpy as np
from os.path import basename

class CustomSubclass(Custom):
    """
    """

    def __init__(
        self,
        x,
        y,
        name = 'custom_subclass',
        ):
        """
        x,y are ints
        """
        Custom.__init__(
            self,
            name = name,
        )

        # attributes which are emdfile classes get saved
        self.x = Array(
            name = 'x',
            data = np.full((2,2),x)
        )
        self.y = PointList(
            name = 'y',
            data = np.full(5,y,dtype=[('a',int),('b',float)])
        )

        pass



    # Read methods

    @classmethod
    def _get_constructor_args(cls,group):
        """
        """
        # get data
        emd_data = cls._get_emd_attr_data(cls,group)

        # get arguments
        constructor_args = {
            'name' : basename(group.name),
            'x' : emd_data['x'],
            'y' : some_metadata['thing']
            # etc.
        }

        # return
        return constructor_args



    # This method is optional

    def _populate_instance(self,group):
        """
        Optional.  During read, this method is run after object instantiation.
        """
        pass




    # Write methods

    # This method is optional

    def to_h5(self,group):
        """
        Optional. If defined, should call Custom.to_h5(self,group) to handle
        saving the parent class' data/metadata.
        """
        # Here some class attributes are stored in a Metadata dict which is
        # placed in the .metadata propetry, such that it's saved in the 
        # subsequent `to_h5` call
        self.metadata = Metadata(
            name = '_attr_metadata',
            data = {
                'x' : self.x,
                'y' : self.y
            }
        )
        Custom.to_h5(self,group)








# This file defines a minimal subclass of Array


from emdfile import Array
import numpy as np
from os.path import basename

class SimpleArraySubclass(Array):
    """
    """

    def __init__(
        self,
        data,
        name = 'a_simple_array',
        **kwargs
        ):
        """
        """
        # Initialize as the parent class
        Array.__init__(
            self,
            name = name,
            data = data
        )

        pass



    # Read methods
    @classmethod
    def _get_constructor_args(cls,group):
        """
        Returns a dictionary of arguments to pass to __init__
        """
        # get array constructor args
        array_constr_args = Array._get_constructor_args(group)

        # return them
        return array_constr_args



